@@ -35,11 +35,6 @@
 
 #include "cvs.h"
 
-#ifndef lint
-static const char rcsid[] = "$CVSid: @(#)checkout.c 1.78 94/10/07 $";
-USE(rcsid)
-#endif
-
 static char *findslash PROTO((char *start, char *p));
 static int build_dirs_and_chdir PROTO((char *dir, char *prepath, char *realdir,
 				 int sticky));
@@ -47,23 +42,22 @@ static int checkout_proc PROTO((int *pargc, char **argv, char *where,
 		          char *mwhere, char *mfile, int shorten,
 		          int local_specified, char *omodule,
 		          char *msg));
+static int safe_location PROTO((void));
 
 static const char *const checkout_usage[] =
 {
-    "Usage:\n  %s %s [-ANPQcflnpqs] [-r rev | -D date] [-d dir] [-k kopt] modules...\n",
+    "Usage:\n  %s %s [-ANPcflnps] [-r rev | -D date] [-d dir] [-k kopt] modules...\n",
     "\t-A\tReset any sticky tags/date/kopts.\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
     "\t-P\tPrune empty directories.\n",
-    "\t-Q\tReally quiet.\n",
     "\t-c\t\"cat\" the module database.\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
     "\t-n\tDo not run module program (if any).\n",
-    "\t-p\tCheck out files to standard output.\n",
-    "\t-q\tSomewhat quiet.\n",
+    "\t-p\tCheck out files to standard output (avoids stickiness).\n",
     "\t-s\tLike -c, but include module status.\n",
-    "\t-r rev\tCheck out revision or tag. (implies -P)\n",
-    "\t-D date\tCheck out revisions as of date. (implies -P)\n",
+    "\t-r rev\tCheck out revision or tag. (implies -P) (is sticky)\n",
+    "\t-D date\tCheck out revisions as of date. (implies -P) (is sticky)\n",
     "\t-d dir\tCheck out into dir instead of module name.\n",
     "\t-k kopt\tUse RCS kopt -k option on checkout.\n",
     "\t-j rev\tMerge in changes made between current revision and rev.\n",
@@ -72,16 +66,15 @@ static const char *const checkout_usage[] =
 
 static const char *const export_usage[] =
 {
-    "Usage: %s %s [-NPQflnq] [-r rev | -D date] [-d dir] module...\n",
+    "Usage: %s %s [-NPfln] [-r rev | -D date] [-d dir] [-k kopt] module...\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
-    "\t-Q\tReally quiet.\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
     "\t-n\tDo not run module program (if any).\n",
-    "\t-q\tSomewhat quiet.\n",
     "\t-r rev\tCheck out revision or tag.\n",
     "\t-D date\tCheck out revisions as of date.\n",
     "\t-d dir\tCheck out into dir instead of module name.\n",
+    "\t-k kopt\tUse RCS kopt -k option on checkout.\n",
     NULL
 };
 
@@ -91,10 +84,13 @@ static int pipeout;
 static int aflag;
 static char *options = NULL;
 static char *tag = NULL;
+static int tag_validated = 0;
 static char *date = NULL;
 static char *join_rev1 = NULL;
 static char *join_rev2 = NULL;
+static int join_tags_validated = 0;
 static char *preload_update_dir = NULL;
+static char *history_name = NULL;
 
 int
 checkout (argc, argv)
@@ -111,20 +107,23 @@ checkout (argc, argv)
     char *where = NULL;
     char *valid_options;
     const char *const *valid_usage;
+    enum mtype m_type;
 
     /*
      * A smaller subset of options are allowed for the export command, which
      * is essentially like checkout, except that it hard-codes certain
-     * options to be on (like -kv) and takes care to remove the CVS directory
-     * when it has done its duty
+     * options to be default (like -kv) and takes care to remove the CVS
+     * directory when it has done its duty
      */
     if (strcmp (command_name, "export") == 0)
     {
-	valid_options = "Nnd:flRQqr:D:";
+        m_type = EXPORT;
+	valid_options = "Nnk:d:flRQqr:D:";
 	valid_usage = export_usage;
     }
     else
     {
+        m_type = CHECKOUT;
 	valid_options = "ANnk:d:flRpQqcsr:D:j:P";
 	valid_usage = checkout_usage;
     }
@@ -133,6 +132,7 @@ checkout (argc, argv)
 	usage (valid_usage);
 
     ign_setup ();
+    wrap_setup ();
 
     optind = 1;
     while ((c = getopt (argc, argv, valid_options)) != -1)
@@ -154,10 +154,15 @@ checkout (argc, argv)
 		run_module_prog = 0;
 		break;
 	    case 'Q':
-		really_quiet = 1;
-		/* FALL THROUGH */
 	    case 'q':
-		quiet = 1;
+#ifdef SERVER_SUPPORT
+		/* The CVS 1.5 client sends these options (in addition to
+		   Global_option requests), so we must ignore them.  */
+		if (!server_active)
+#endif
+		    error (1, 0,
+			   "-q or -Q must be specified before \"%s\"",
+			   command_name);
 		break;
 	    case 'l':
 		local = 1;
@@ -215,8 +220,7 @@ checkout (argc, argv)
     if (shorten == -1)
 	shorten = 0;
 
-    if ((!(cat + status) && argc == 0) || ((cat + status) && argc != 0)
-	|| (tag && date))
+    if ((!(cat + status) && argc == 0) || ((cat + status) && argc != 0))
 	usage (valid_usage);
 
     if (where && pipeout)
@@ -231,9 +235,21 @@ checkout (argc, argv)
 	}
 	if (tag && isdigit (tag[0]))
 	    error (1, 0, "tag `%s' must be a symbolic tag", tag);
-	options = RCS_check_kflag ("v");/* -kv must be on */
+/*
+ * mhy 950615: -kv doesn't work for binaries with RCS keywords.
+ * Instead use the default provided in the RCS file (-ko for binaries).
+ */
+#if 0
+	if (!options)
+	  options = RCS_check_kflag ("v");/* -kv is default */
+#endif
     }
 
+    if (!safe_location()) {
+        error(1, 0, "Cannot check out files into the repository itself");
+    }
+
+#ifdef CLIENT_SUPPORT
     if (client_active)
     {
 	int expand_modules;
@@ -254,17 +270,15 @@ checkout (argc, argv)
 			  && supported_request ("expand-modules"));
 	
 	if (expand_modules)
-	  {
+	{
 	    /* This is done here because we need to read responses
                from the server before we send the command checkout or
                export files. */
 
 	    client_expand_modules (argc, argv, local);
-	  }
+	}
 
 	if (!run_module_prog) send_arg ("-n");
-	if (really_quiet) send_arg ("-Q");
-	if (quiet) send_arg ("-q");
 	if (local) send_arg ("-l");
 	if (pipeout) send_arg ("-p");
 	if (!force_tag_match) send_arg ("-f");
@@ -296,25 +310,24 @@ checkout (argc, argv)
 	    option_with_arg ("-j", join_rev2);
 
 	if (expand_modules)
-	  {
-	    client_send_expansions (local);
-	  }
+	{
+	    client_send_expansions (local, where);
+	}
 	else
-	  {
+	{
 	    int i;
 	    for (i = 0; i < argc; ++i)
-	      send_arg (argv[i]);
+		send_arg (argv[i]);
 	    client_nonexpanded_setup ();
-	  }
+	}
 
-	if (fprintf
-	    (to_server,
-	     strcmp (command_name, "export") == 0 ? "export\n" : "co\n")
-	    < 0)
-	  error (1, errno, "writing to server");
+	send_to_server (strcmp (command_name, "export") == 0 ?
+                        "export\012" : "co\012",
+                        0);
 
 	return get_responses_and_close ();
     }
+#endif /* CLIENT_SUPPORT */
 
     if (cat || status)
     {
@@ -332,23 +345,28 @@ checkout (argc, argv)
     {
 	char repository[PATH_MAX];
 
-	(void) mkdir (where, 0777);
-	if (chdir (where) < 0)
+	(void) CVS_MKDIR (where, 0777);
+	if ( CVS_CHDIR (where) < 0)
 	    error (1, errno, "cannot chdir to %s", where);
 	preload_update_dir = xstrdup (where);
 	where = (char *) NULL;
 	if (!isfile (CVSADM))
 	{
-	    (void) sprintf (repository, "%s/%s/%s", CVSroot, CVSROOTADM,
-			    CVSNULLREPOS);
+	    (void) sprintf (repository, "%s/%s/%s", CVSroot_directory,
+			    CVSROOTADM, CVSNULLREPOS);
 	    if (!isfile (repository))
-		(void) mkdir (repository, 0777);
+	    {
+		mode_t omask;
+		omask = umask (cvsumask);
+		(void) CVS_MKDIR (repository, 0777);
+		(void) umask (omask);
+	    }
 
 	    /* I'm not sure whether this check is redundant.  */
 	    if (!isdir (repository))
 		error (1, 0, "there is no repository %s", repository);
 
-	    Create_Admin (".", where, repository,
+	    Create_Admin (".", preload_update_dir, repository,
 			  (char *) NULL, (char *) NULL);
 	    if (!noexec)
 	    {
@@ -357,12 +375,29 @@ checkout (argc, argv)
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+#ifdef SERVER_SUPPORT
 		if (server_active)
 		    server_set_entstat (preload_update_dir, repository);
+#endif
 	    }
 	}
     }
 
+    /* If we will be calling history_write, work out the name to pass
+       it.  */
+    if (strcmp (command_name, "export") != 0 && !pipeout)
+    {
+	if (tag && date)
+	{
+	    history_name = xmalloc (strlen (tag) + strlen (date) + 2);
+	    sprintf (history_name, "%s:%s", tag, date);
+	}
+	else if (tag)
+	    history_name = tag;
+	else
+	    history_name = date;
+    }
+
     /*
      * if where was specified (-d) and we have not taken care of it already
      * with the multiple arg stuff, and it was not a simple directory name
@@ -377,7 +412,7 @@ checkout (argc, argv)
 	slash = strrchr (where, '/');
 	*slash = '\0';
 
-	if (chdir (where) < 0)
+	if ( CVS_CHDIR (where) < 0)
 	    error (1, errno, "cannot chdir to %s", where);
 
 	preload_update_dir = xstrdup (where);
@@ -388,13 +423,48 @@ checkout (argc, argv)
     }
 
     for (i = 0; i < argc; i++)
-	err += do_module (db, argv[i], CHECKOUT, "Updating", checkout_proc,
+	err += do_module (db, argv[i], m_type, "Updating", checkout_proc,
 			  where, shorten, local, run_module_prog,
 			  (char *) NULL);
     close_module (db);
     return (err);
 }
 
+static int
+safe_location ()
+{
+    char current[PATH_MAX];
+    char hardpath[PATH_MAX+5];
+    size_t hardpath_len;
+    int  x;
+
+    x = readlink(CVSroot_directory, hardpath, sizeof hardpath - 1);
+    if (x == -1)
+    {
+        strcpy(hardpath, CVSroot_directory);
+    }
+    else
+    {
+        hardpath[x] = '\0';
+    }
+    getwd (current);
+    hardpath_len = strlen (hardpath);
+    if (strncmp (current, hardpath, hardpath_len) == 0)
+    {
+	if (/* Current is a subdirectory of hardpath.  */
+	    current[hardpath_len] == '/'
+
+	    /* Current is hardpath itself.  */
+	    || current[hardpath_len] == '\0')
+	    return 0;
+	else
+	    /* It isn't a problem.  For example, current is
+	       "/foo/cvsroot-bar" and hardpath is "/foo/cvsroot".  */
+	    return 1;
+    }
+    return (1);
+}
+
 /*
  * process_module calls us back here so we do the actual checkout stuff
  */
@@ -433,7 +503,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
      */
 
     /* set up the repository (maybe) for the bottom directory */
-    (void) sprintf (repository, "%s/%s", CVSroot, argv[0]);
+    (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
 
     /* save the original value of preload_update_dir */
     if (preload_update_dir != NULL)
@@ -624,13 +694,15 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		if (!isdir (repository))
 		    error (1, 0, "there is no repository %s", repository);
 
-		Create_Admin (".", where, repository,
+		Create_Admin (".", preload_update_dir, repository,
 			      (char *) NULL, (char *) NULL);
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+#ifdef SERVER_SUPPORT
 		if (server_active)
 		    server_set_entstat (where, repository);
+#endif
 	    }
 	    else
 	    {
@@ -638,7 +710,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		if (!isdir (repository))
 		    error (1, 0, "there is no repository %s", repository);
 
-		Create_Admin (".", where, repository, tag, date);
+		Create_Admin (".", preload_update_dir, repository, tag, date);
 	    }
 	}
 	else
@@ -647,7 +719,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 
 	    /* get the contents of the previously existing repository */
 	    repos = Name_Repository ((char *) NULL, preload_update_dir);
-	    if (strcmp (repository, repos) != 0)
+	    if (fncmp (repository, repos) != 0)
 	    {
 		error (0, 0, "existing repository %s does not match %s",
 		       repos, repository);
@@ -668,7 +740,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
      */
     if (pipeout)
     {
-	if (chdir (repository) < 0)
+	if ( CVS_CHDIR (repository) < 0)
 	{
 	    error (0, errno, "cannot chdir to %s", repository);
 	    free (preload_update_dir);
@@ -676,13 +748,37 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	    return (1);
 	}
 	which = W_REPOS;
+	if (tag != NULL && !tag_validated)
+	{
+	    tag_check_valid (tag, *pargc - 1, argv + 1, 0, aflag, NULL);
+	    tag_validated = 1;
+	}
     }
     else
+    {
 	which = W_LOCAL | W_REPOS;
+	if (tag != NULL && !tag_validated)
+	{
+	    tag_check_valid (tag, *pargc - 1, argv + 1, 0, aflag,
+			     repository);
+	    tag_validated = 1;
+	}
+    }
 
-    if (tag != NULL || date != NULL)
+    if (tag != NULL || date != NULL || join_rev1 != NULL)
 	which |= W_ATTIC;
 
+    if (! join_tags_validated)
+    {
+        if (join_rev1 != NULL)
+	    tag_check_valid_join (join_rev1, *pargc - 1, argv + 1, 0, aflag,
+				  repository);
+	if (join_rev2 != NULL)
+	    tag_check_valid_join (join_rev2, *pargc - 1, argv + 1, 0, aflag,
+				  repository);
+	join_tags_validated = 1;
+    }
+
     /*
      * if we are going to be recursive (building dirs), go ahead and call the
      * update recursion processor.  We will be recursive unless either local
@@ -691,7 +787,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
     if (!(local_specified || *pargc > 1))
     {
 	if (strcmp (command_name, "export") != 0 && !pipeout)
-	    history_write ('O', preload_update_dir, tag ? tag : date, where,
+	    history_write ('O', preload_update_dir, history_name, where,
 			   repository);
 	err += do_update (0, (char **) NULL, options, tag, date,
 			  force_tag_match, 0 /* !local */ ,
@@ -712,19 +808,32 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	entries = Entries_Open (0);
 	for (i = 1; i < *pargc; i++)
 	{
-	    char line[MAXLINELEN];
-	    char *user;
+	    char *line;
 	    Vers_TS *vers;
+	    struct file_info finfo;
 
-	    user = argv[i];
-	    vers = Version_TS (repository, options, tag, date, user,
-			       force_tag_match, 0, entries, (List *) NULL);
+	    memset (&finfo, 0, sizeof finfo);
+	    finfo.file = argv[i];
+	    /* Shouldn't be used, so set to arbitrary value.  */
+	    finfo.update_dir = NULL;
+	    finfo.fullname = argv[i];
+	    finfo.repository = repository;
+	    finfo.entries = entries;
+	    /* The rcs slot is needed to get the options from the RCS
+               file */
+	    finfo.rcs = RCS_parse (finfo.file, repository);
+
+	    vers = Version_TS (&finfo, options, tag, date,
+			       force_tag_match, 0);
 	    if (vers->ts_user == NULL)
 	    {
-		(void) sprintf (line, "Initial %s", user);
-		Register (entries, user, vers->vn_rcs ? vers->vn_rcs : "0",
+		line = xmalloc (strlen (finfo.file) + 15);
+		(void) sprintf (line, "Initial %s", finfo.file);
+		Register (entries, finfo.file,
+			  vers->vn_rcs ? vers->vn_rcs : "0",
 			  line, vers->options, vers->tag,
 			  vers->date, (char *) 0);
+		free (line);
 	    }
 	    freevers_ts (&vers);
 	}
@@ -734,7 +843,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 
     /* Don't log "export", just regular "checkouts" */
     if (strcmp (command_name, "export") != 0 && !pipeout)
-	history_write ('O', preload_update_dir, (tag ? tag : date), where,
+	history_write ('O', preload_update_dir, history_name, where,
 		       repository);
 
     /* go ahead and call update now that everything is set */
@@ -788,8 +897,9 @@ build_dirs_and_chdir (dir, prepath, realdir, sticky)
     {
 	*slash = '\0';
 	*slash2 = '\0';
-	(void) mkdir (cp, 0777);
-	if (chdir (cp) < 0)
+	Subdir_Register ((List *) NULL, (char *) NULL, cp);
+	(void) CVS_MKDIR (cp, 0777);
+	if ( CVS_CHDIR (cp) < 0)
 	{
 	    error (0, errno, "cannot chdir to %s", cp);
 	    return (1);
@@ -800,22 +910,25 @@ build_dirs_and_chdir (dir, prepath, realdir, sticky)
 	    /* I'm not sure whether this check is redundant.  */
 	    if (!isdir (repository))
 		error (1, 0, "there is no repository %s", repository);
-	    Create_Admin (".", cp, repository, sticky ? (char *) NULL : tag,
+	    Create_Admin (".", path, repository, sticky ? (char *) NULL : tag,
 			  sticky ? (char *) NULL : date);
 	    if (!noexec)
 	    {
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
+#ifdef SERVER_SUPPORT
 		if (server_active)
 		    server_set_entstat (path, repository);
+#endif
 	    }
 	}
 	*slash = '/';
 	*slash2 = '/';
     }
-    (void) mkdir (cp, 0777);
-    if (chdir (cp) < 0)
+    Subdir_Register ((List *) NULL, (char *) NULL, cp);
+    (void) CVS_MKDIR (cp, 0777);
+    if ( CVS_CHDIR (cp) < 0)
     {
 	error (0, errno, "cannot chdir to %s", cp);
 	return (1);
