@@ -13,15 +13,12 @@
  * size, and this code works fine.
  */
 
+#include <assert.h>
 #include "cvs.h"
+#include "getline.h"
 
 #ifdef MY_NDBM
 
-#ifndef lint
-static const char rcsid[] = "$CVSid: @(#)myndbm.c 1.7 94/09/23 $";
-USE(rcsid)
-#endif
-
 static void mydbm_load_file ();
 
 /* ARGSUSED */
@@ -34,21 +31,54 @@ mydbm_open (file, flags, mode)
     FILE *fp;
     DBM *db;
 
-    if ((fp = fopen (file, "r")) == NULL)
+    fp = CVS_FOPEN (file, FOPEN_BINARY_READ);
+    if (fp == NULL && !(existence_error (errno) && (flags & O_CREAT)))
 	return ((DBM *) 0);
 
     db = (DBM *) xmalloc (sizeof (*db));
     db->dbm_list = getlist ();
+    db->modified = 0;
+    db->name = xstrdup (file);
 
-    mydbm_load_file (fp, db->dbm_list);
-    (void) fclose (fp);
+    if (fp != NULL)
+    {
+	mydbm_load_file (fp, db->dbm_list);
+	if (fclose (fp) < 0)
+	    error (0, errno, "cannot close %s", file);
+    }
     return (db);
 }
 
+static int write_item PROTO ((Node *, void *));
+
+static int
+write_item (node, data)
+    Node *node;
+    void *data;
+{
+    FILE *fp = (FILE *)data;
+    fputs (node->key, fp);
+    fputs (" ", fp);
+    fputs (node->data, fp);
+    fputs ("\012", fp);
+    return 0;
+}
+
 void
 mydbm_close (db)
     DBM *db;
 {
+    if (db->modified)
+    {
+	FILE *fp;
+	fp = CVS_FOPEN (db->name, FOPEN_BINARY_WRITE);
+	if (fp == NULL)
+	    error (1, errno, "cannot write %s", db->name);
+	walklist (db->dbm_list, write_item, (void *)fp);
+	if (fclose (fp) < 0)
+	    error (0, errno, "cannot close %s", db->name);
+    }
+    free (db->name);
     dellist (&db->dbm_list);
     free ((char *) db);
 }
@@ -128,19 +158,63 @@ mydbm_nextkey (db)
     return (key);
 }
 
+/* Note: only updates the in-memory copy, which is written out at
+   mydbm_close time.  Note: Also differs from DBM in that on duplication,
+   it gives a warning, rather than either DBM_INSERT or DBM_REPLACE
+   behavior.  */
+int
+mydbm_store (db, key, value, flags)
+    DBM *db;
+    datum key;
+    datum value;
+    int flags;
+{
+    Node *node;
+
+    node = getnode ();
+    node->type = NDBMNODE;
+
+    node->key = xmalloc (key.dsize + 1);
+    strncpy (node->key, key.dptr, key.dsize);
+    node->key[key.dsize] = '\0';
+
+    node->data = xmalloc (value.dsize + 1);
+    strncpy (node->data, value.dptr, value.dsize);
+    node->data[value.dsize] = '\0';
+
+    db->modified = 1;
+    if (addnode (db->dbm_list, node) == -1)
+    {
+	error (0, 0, "attempt to insert duplicate key `%s'", node->key);
+	freenode (node);
+	return 0;
+    }
+    return 0;
+}
+
 static void
 mydbm_load_file (fp, list)
     FILE *fp;
     List *list;
 {
-    char line[MAXLINELEN], value[MAXLINELEN];
+    char *line = NULL;
+    size_t line_len;
+    /* FIXME: arbitrary limit.  */
+    char value[MAXLINELEN];
     char *cp, *vp;
     int len, cont;
 
-    for (cont = 0; fgets (line, sizeof (line), fp) != NULL;)
+    for (cont = 0; getline (&line, &line_len, fp) >= 0;)
     {
-	if ((cp = strrchr (line, '\n')) != NULL)
+	if ((cp = strrchr (line, '\012')) != NULL)
 	    *cp = '\0';			/* strip the newline */
+	cp = line + strlen (line);
+	if (cp > line && cp[-1] == '\015')
+	    /* If the file (e.g. modules) was written on an NT box, it will
+	       contain CRLF at the ends of lines.  Strip them (we can't do
+	       this by opening the file in text mode because we might be
+	       running on unix).  */
+	    cp[-1] = '\0';	    
 
 	/*
 	 * Add the line to the value, at the end if this is a continuation
@@ -208,6 +282,7 @@ mydbm_load_file (fp, list)
 	    }
 	}
     }
+    free (line);
 }
 
 #endif				/* MY_NDBM */
