@@ -16,8 +16,10 @@
 #include "getline.h"
 
 static RETSIGTYPE patch_cleanup PROTO((void));
-static Dtype patch_dirproc PROTO((char *dir, char *repos, char *update_dir));
-static int patch_fileproc PROTO((struct file_info *finfo));
+static Dtype patch_dirproc PROTO ((void *callerdat, char *dir,
+				   char *repos, char *update_dir,
+				   List *entries));
+static int patch_fileproc PROTO ((void *callerdat, struct file_info *finfo));
 static int patch_proc PROTO((int *pargc, char **argv, char *xwhere,
 		       char *mwhere, char *mfile, int shorten,
 		       int local_specified, char *mname, char *msg));
@@ -28,12 +30,14 @@ static int toptwo_diffs = 0;
 static int local = 0;
 static char *options = NULL;
 static char *rev1 = NULL;
-static int rev1_validated = 1;
+static int rev1_validated = 0;
 static char *rev2 = NULL;
-static int rev2_validated = 1;
+static int rev2_validated = 0;
 static char *date1 = NULL;
 static char *date2 = NULL;
-static char tmpfile1[L_tmpnam+1], tmpfile2[L_tmpnam+1], tmpfile3[L_tmpnam+1];
+static char *tmpfile1 = NULL;
+static char *tmpfile2 = NULL;
+static char *tmpfile3 = NULL;
 static int unidiff = 0;
 
 static const char *const patch_usage[] =
@@ -253,7 +257,7 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
     int which;
     char repository[PATH_MAX];
 
-    (void) sprintf (repository, "%s/%s", CVSroot, argv[0]);
+    (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
     (void) strcpy (where, argv[0]);
 
     /* if mfile isn't null, we need to set up to do only part of the module */
@@ -295,7 +299,7 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
     }
 
     /* cd to the starting repository */
-    if (chdir (repository) < 0)
+    if ( CVS_CHDIR (repository) < 0)
     {
 	error (0, errno, "cannot chdir to %s", repository);
 	return (1);
@@ -319,8 +323,9 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 
     /* start the recursion processor */
     err = start_recursion (patch_fileproc, (FILESDONEPROC) NULL, patch_dirproc,
-			   (DIRLEAVEPROC) NULL, *pargc - 1, argv + 1, local,
-			   which, 0, 1, where, 1, 1);
+			   (DIRLEAVEPROC) NULL, NULL,
+			   *pargc - 1, argv + 1, local,
+			   which, 0, 1, where, 1);
 
     return (err);
 }
@@ -331,7 +336,8 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
  */
 /* ARGSUSED */
 static int
-patch_fileproc (finfo)
+patch_fileproc (callerdat, finfo)
+    void *callerdat;
     struct file_info *finfo;
 {
     struct utimbuf t;
@@ -362,7 +368,15 @@ patch_fileproc (finfo)
     if (isattic && rev2 == NULL && date2 == NULL)
 	vers_head = NULL;
     else
-	vers_head = RCS_getversion (rcsfile, rev2, date2, force_tag_match, 0);
+    {
+	vers_head = RCS_getversion (rcsfile, rev2, date2, force_tag_match,
+				    (int *) NULL);
+	if (vers_head != NULL && RCS_isdead (rcsfile, vers_head))
+	{
+	    free (vers_head);
+	    vers_head = NULL;
+	}
+    }
 
     if (toptwo_diffs)
     {
@@ -380,9 +394,15 @@ patch_fileproc (finfo)
 	    return (1);
 	}
     }
-    vers_tag = RCS_getversion (rcsfile, rev1, date1, force_tag_match, 0);
+    vers_tag = RCS_getversion (rcsfile, rev1, date1, force_tag_match,
+			       (int *) NULL);
+    if (vers_tag != NULL && RCS_isdead (rcsfile, vers_tag))
+    {
+        free (vers_tag);
+	vers_tag = NULL;
+    }
 
-    if (vers_tag == NULL && (vers_head == NULL || isattic))
+    if (vers_tag == NULL && vers_head == NULL)
 	return (0);			/* nothing known about specified revs */
 
     if (vers_tag && vers_head && strcmp (vers_head, vers_tag) == 0)
@@ -409,22 +429,27 @@ patch_fileproc (finfo)
 			   vers_tag, vers_head);
 	return (0);
     }
-    if ((fp1 = fopen (tmpnam (tmpfile1), "w+")) != NULL)
+    tmpfile1 = cvs_temp_name ();
+    if ((fp1 = CVS_FOPEN (tmpfile1, "w+")) != NULL)
 	(void) fclose (fp1);
-    if ((fp2 = fopen (tmpnam (tmpfile2), "w+")) != NULL)
+    tmpfile2 = cvs_temp_name ();
+    if ((fp2 = CVS_FOPEN (tmpfile2, "w+")) != NULL)
 	(void) fclose (fp2);
-    if ((fp3 = fopen (tmpnam (tmpfile3), "w+")) != NULL)
+    tmpfile3 = cvs_temp_name ();
+    if ((fp3 = CVS_FOPEN (tmpfile3, "w+")) != NULL)
 	(void) fclose (fp3);
     if (fp1 == NULL || fp2 == NULL || fp3 == NULL)
     {
+	/* FIXME: should be printing a proper error message, with errno-based
+	   message, and the filename which we could not create.  */
 	error (0, 0, "cannot create temporary files");
 	ret = 1;
 	goto out;
     }
     if (vers_tag != NULL)
     {
-	retcode = RCS_checkout (rcsfile->path, NULL, vers_tag, options, tmpfile1,
-	                        0, 0);
+	retcode = RCS_checkout (rcsfile, (char *) NULL, vers_tag,
+				(char *) NULL, options, tmpfile1);
 	if (retcode != 0)
 	{
 	    if (!really_quiet)
@@ -445,7 +470,8 @@ patch_fileproc (finfo)
     }
     if (vers_head != NULL)
     {
-	retcode = RCS_checkout (rcsfile->path, NULL, vers_head, options, tmpfile2, 0, 0);
+	retcode = RCS_checkout (rcsfile, (char *) NULL, vers_head,
+				(char *) NULL, options, tmpfile2);
 	if (retcode != 0)
 	{
 	    if (!really_quiet)
@@ -522,8 +548,8 @@ patch_fileproc (finfo)
 		    goto out;
 		}
 	    }
-	    if (CVSroot != NULL)
-		(void) sprintf (strippath, "%s/", CVSroot);
+	    if (CVSroot_directory != NULL)
+		(void) sprintf (strippath, "%s/", CVSroot_directory);
 	    else
 		(void) strcpy (strippath, REPOS_STRIP);
 	    if (strncmp (rcs, strippath, strlen (strippath)) == 0)
@@ -568,9 +594,13 @@ patch_fileproc (finfo)
     if (line2)
         free (line2);
     /* FIXME: should be checking for errors.  */
-    (void) unlink (tmpfile1);
-    (void) unlink (tmpfile2);
-    (void) unlink (tmpfile3);
+    (void) CVS_UNLINK (tmpfile1);
+    (void) CVS_UNLINK (tmpfile2);
+    (void) CVS_UNLINK (tmpfile3);
+    free (tmpfile1);
+    free (tmpfile2);
+    free (tmpfile3);
+    tmpfile1 = tmpfile2 = tmpfile3 = NULL;
     return (ret);
 }
 
@@ -579,10 +609,12 @@ patch_fileproc (finfo)
  */
 /* ARGSUSED */
 static Dtype
-patch_dirproc (dir, repos, update_dir)
+patch_dirproc (callerdat, dir, repos, update_dir, entries)
+    void *callerdat;
     char *dir;
     char *repos;
     char *update_dir;
+    List *entries;
 {
     if (!quiet)
 	error (0, 0, "Diffing %s", update_dir);
@@ -595,10 +627,20 @@ patch_dirproc (dir, repos, update_dir)
 static RETSIGTYPE
 patch_cleanup ()
 {
-    if (tmpfile1[0] != '\0')
+    if (tmpfile1 != NULL)
+    {
 	(void) unlink_file (tmpfile1);
-    if (tmpfile2[0] != '\0')
+	free (tmpfile1);
+    }
+    if (tmpfile2 != NULL)
+    {
 	(void) unlink_file (tmpfile2);
-    if (tmpfile3[0] != '\0')
+	free (tmpfile2);
+    }
+    if (tmpfile3 != NULL)
+    {
 	(void) unlink_file (tmpfile3);
+	free (tmpfile3);
+    }
+    tmpfile1 = tmpfile2 = tmpfile3 = NULL;
 }
