@@ -27,7 +27,6 @@ Name_Repository (dir, update_dir)
     char repos[PATH_MAX];
     char path[PATH_MAX];
     char tmp[PATH_MAX];
-    char cvsadm[PATH_MAX];
     char *cp;
 
     if (update_dir && *update_dir)
@@ -35,46 +34,42 @@ Name_Repository (dir, update_dir)
     else
 	xupdate_dir = ".";
 
-    if (dir != NULL)
-	(void) sprintf (cvsadm, "%s/%s", dir, CVSADM);
-    else
-	(void) strcpy (cvsadm, CVSADM);
-
-    /* sanity checks */
-    if (!isdir (cvsadm))
-    {
-	error (0, 0, "in directory %s:", xupdate_dir);
-	error (1, 0, "there is no version here; do '%s checkout' first",
-	       program_name);
-    }
-
-    if (dir != NULL)
-	(void) sprintf (tmp, "%s/%s", dir, CVSADM_ENT);
-    else
-	(void) strcpy (tmp, CVSADM_ENT);
-
-    if (!isreadable (tmp))
-    {
-	error (0, 0, "in directory %s:", xupdate_dir);
-	error (1, 0, "*PANIC* administration files missing");
-    }
-
     if (dir != NULL)
 	(void) sprintf (tmp, "%s/%s", dir, CVSADM_REP);
     else
 	(void) strcpy (tmp, CVSADM_REP);
 
-    if (!isreadable (tmp))
-    {
-	error (0, 0, "in directory %s:", xupdate_dir);
-	error (1, 0, "*PANIC* administration files missing");
-    }
-
     /*
      * The assumption here is that the repository is always contained in the
      * first line of the "Repository" file.
      */
-    fpin = open_file (tmp, "r");
+    fpin = CVS_FOPEN (tmp, "r");
+
+    if (fpin == NULL)
+    {
+	int save_errno = errno;
+	char cvsadm[PATH_MAX];
+
+	if (dir != NULL)
+	    (void) sprintf (cvsadm, "%s/%s", dir, CVSADM);
+	else
+	    (void) strcpy (cvsadm, CVSADM);
+
+	if (!isdir (cvsadm))
+	{
+	    error (0, 0, "in directory %s:", xupdate_dir);
+	    error (1, 0, "there is no version here; do '%s checkout' first",
+		   program_name);
+	}
+
+	if (existence_error (save_errno))
+	{
+	    error (0, 0, "in directory %s:", xupdate_dir);
+	    error (1, 0, "*PANIC* administration files missing");
+	}
+
+	error (1, save_errno, "cannot open %s", tmp);
+    }
 
     if (fgets (repos, PATH_MAX, fpin) == NULL)
     {
@@ -98,7 +93,7 @@ Name_Repository (dir, update_dir)
     }
     if (! isabsolute(repos))
     {
-	if (CVSroot == NULL)
+	if (CVSroot_original == NULL)
 	{
 	    error (0, 0, "in directory %s:", xupdate_dir);
 	    error (0, 0, "must set the CVSROOT environment variable\n");
@@ -106,16 +101,7 @@ Name_Repository (dir, update_dir)
 	    error (1, 0, "illegal repository setting");
 	}
 	(void) strcpy (path, repos);
-	(void) sprintf (repos, "%s/%s", CVSroot, path);
-    }
-#ifdef CLIENT_SUPPORT
-    if (!client_active && !isdir (repos))
-#else
-    if (!isdir (repos))
-#endif
-    {
-	error (0, 0, "in directory %s:", xupdate_dir);
-	error (1, 0, "there is no repository %s", repos);
+	(void) sprintf (repos, "%s/%s", CVSroot_directory, path);
     }
 
     /* allocate space to return and fill it in */
@@ -137,9 +123,10 @@ Short_Repository (repository)
 
     /* If repository matches CVSroot at the beginning, strip off CVSroot */
     /* And skip leading '/' in rep, in case CVSroot ended with '/'. */
-    if (strncmp (CVSroot, repository, strlen (CVSroot)) == 0)
+    if (strncmp (CVSroot_directory, repository,
+		 strlen (CVSroot_directory)) == 0)
     {
-	char *rep = repository + strlen (CVSroot);
+	char *rep = repository + strlen (CVSroot_directory);
 	return (*rep == '/') ? rep+1 : rep;
     }
     else
